use crate::{
    config::ImageFormat,
    error::{Error, UploadError},
    ingest::Session,
    queue::{Base64Bytes, LocalBoxFuture, Process},
    repo::{Alias, DeleteToken, FullRepo, UploadId, UploadResult},
    serde_str::Serde,
    store::{Identifier, Store},
};
use futures_util::{StreamExt, TryStreamExt};
use reqwest_middleware::ClientWithMiddleware;
use std::path::PathBuf;
use tracing::Instrument;
use url::Url;

pub(super) fn perform<'a, R, S>(
    repo: &'a R,
    store: &'a S,
    client: &'a ClientWithMiddleware,
    job: &'a [u8],
) -> LocalBoxFuture<'a, Result<(), Error>>
where
    R: FullRepo + 'static,
    S: Store + 'static,
{
    Box::pin(async move {
        match serde_json::from_slice(job) {
            Ok(job) => match job {
                Process::Ingest {
                    identifier: Base64Bytes(identifier),
                    upload_id,
                    declared_alias,
                    should_validate,
                } => {
                    process_ingest(
                        repo,
                        store,
                        client,
                        identifier,
                        Serde::into_inner(upload_id),
                        declared_alias.map(Serde::into_inner),
                        crate::CONFIG.media.external_validation.as_ref(),
                        crate::CONFIG.media.external_validation_timeout,
                        should_validate,
                        crate::CONFIG.media.process_timeout,
                    )
                    .await?
                }
                Process::Generate {
                    target_format,
                    source,
                    process_path,
                    process_args,
                } => {
                    generate(
                        repo,
                        store,
                        target_format,
                        Serde::into_inner(source),
                        process_path,
                        process_args,
                        crate::CONFIG.media.process_timeout,
                    )
                    .await?
                }
                Process::PrepareUpgrade => prepare_upgrade(repo, store).await?,
            },
            Err(e) => {
                tracing::warn!("Invalid job: {}", format!("{e}"));
            }
        }

        Ok(())
    })
}

struct LogDroppedIngest {
    armed: bool,
    upload_id: UploadId,
}

impl LogDroppedIngest {
    fn guard(upload_id: UploadId) -> Self {
        Self {
            armed: true,
            upload_id,
        }
    }

    fn disarm(mut self) {
        self.armed = false;
    }
}

impl Drop for LogDroppedIngest {
    fn drop(&mut self) {
        if self.armed {
            tracing::warn!(
                "Failed to complete an upload {}- clients will hang",
                self.upload_id
            );
        }
    }
}

#[allow(clippy::too_many_arguments)]
#[tracing::instrument(skip_all)]
async fn process_ingest<R, S>(
    repo: &R,
    store: &S,
    client: &ClientWithMiddleware,
    unprocessed_identifier: Vec<u8>,
    upload_id: UploadId,
    declared_alias: Option<Alias>,
    external_validation: Option<&Url>,
    external_validation_timeout: u64,
    should_validate: bool,
    timeout: u64,
) -> Result<(), Error>
where
    R: FullRepo + 'static,
    S: Store + 'static,
{
    let guard = LogDroppedIngest::guard(upload_id);

    let fut = async {
        let unprocessed_identifier = S::Identifier::from_bytes(unprocessed_identifier)?;

        let ident = unprocessed_identifier.clone();

        let store2 = store.clone();
        let repo = repo.clone();
        let client = client.clone();
        let external_validation = external_validation.cloned();
        let error_boundary = actix_rt::spawn(async move {
            let stream = store2
                .to_stream(&ident, None, None)
                .await?
                .map_err(Error::from);

            let session = crate::ingest::ingest(
                &repo,
                &store2,
                &client,
                stream,
                declared_alias,
                external_validation.as_ref(),
                external_validation_timeout,
                should_validate,
                timeout,
            )
            .await?;

            let token = session.delete_token().await?;

            Ok((session, token)) as Result<(Session<R, S>, DeleteToken), Error>
        })
        .await;

        store.remove(&unprocessed_identifier).await?;

        error_boundary.map_err(|_| UploadError::Canceled)?
    };

    let result = match fut.await {
        Ok((mut session, token)) => {
            let alias = session.alias().take().expect("Alias should exist").clone();
            let result = UploadResult::Success { alias, token };
            session.disarm();
            result
        }
        Err(e) => {
            tracing::warn!("Failed to ingest\n{}\n{}", format!("{e}"), format!("{e:?}"));

            UploadResult::Failure {
                message: e.root_cause().to_string(),
            }
        }
    };

    repo.complete(upload_id, result).await?;

    guard.disarm();

    Ok(())
}

#[tracing::instrument(skip_all)]
async fn generate<R: FullRepo, S: Store + 'static>(
    repo: &R,
    store: &S,
    target_format: ImageFormat,
    source: Alias,
    process_path: PathBuf,
    process_args: Vec<String>,
    timeout: u64,
) -> Result<(), Error> {
    let Some(hash) = repo.hash(&source).await? else {
        // Nothing to do
        return Ok(());
    };

    let path_string = process_path.to_string_lossy().to_string();
    let identifier_opt = repo
        .variant_identifier::<S::Identifier>(hash.clone(), path_string)
        .await?;

    if identifier_opt.is_some() {
        return Ok(());
    }

    let original_details = crate::ensure_details(repo, store, &source).await?;

    crate::generate::generate(
        repo,
        store,
        target_format,
        source,
        process_path,
        process_args,
        original_details.to_input_format(),
        None,
        timeout,
        hash,
    )
    .await?;

    Ok(())
}

#[tracing::instrument(skip_all)]
async fn prepare_upgrade<R: FullRepo + 'static, S: Store + 'static>(
    repo: &R,
    store: &S,
) -> Result<(), Error> {
    repo.set("upgrade-preparations-started", b"1".to_vec().into())
        .await?;

    let mut hashes = std::pin::pin!(repo.hashes().await);

    let mut count: u64 = 0;

    while let Some(res) = hashes.next().await {
        match res {
            Ok(hash) => {
                let repo = repo.clone();
                let store = store.clone();

                let current_span = tracing::Span::current();
                let span = tracing::debug_span!(parent: current_span, "ensure details");

                let res = actix_rt::spawn(
                    async move {
                        if let Some(identifier) = repo.identifier(hash).await? {
                            crate::ensure_details_identifier(&repo, &store, &identifier, None)
                                .await?;
                        }

                        repo.set(
                            "upgrade-preparations-progress",
                            Vec::from(count.to_be_bytes()).into(),
                        )
                        .await?;

                        Ok(()) as Result<(), Error>
                    }
                    .instrument(span),
                )
                .await;

                match res {
                    Ok(Ok(_)) => {}
                    Ok(Err(e)) => {
                        tracing::warn!("Failed to ensure details for a hash: {e:?}");
                    }
                    Err(_) => {
                        tracing::warn!("ensure details panicked");
                    }
                }
            }
            Err(e) => {
                tracing::warn!("Skipping ensure details for a hash: {e:?}");
            }
        }

        count += 1;
    }

    repo.set("upgrade-preparations-complete", b"1".to_vec().into())
        .await?;

    Ok(())
}
